/*+
    ARES/HADES/BORG Package -- ./libLSS/data/lensing_survey.hpp
    Copyright (C) 2020 Natalia Porqueres <n.porqueres@imperial.ac.uk>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021)

+*/
#ifndef __LIBLSS_DATA_LENSING_HPP
#  define __LIBLSS_DATA_LENSING_HPP

#  include <H5Cpp.h>
#  include <boost/utility/base_from_member.hpp>
#  include <boost/mpl/assert.hpp>
#  include <boost/utility/enable_if.hpp>
#  include <boost/multi_array.hpp>
#  include <healpix_cxx/pointing.h>
#  include "libLSS/data/base.hpp"
#  include "libLSS/tools/allocator_policy.hpp"
#  include "libLSS/tools/checkmem.hpp"
#  include "libLSS/physics/cosmo.hpp"
#  include "libLSS/physics/projector_lensing.hpp"

namespace LibLSS {
  template <class GT, class AllocationPolicy = DefaultAllocationPolicy>
  class LensingSurvey : virtual LibLSS::Base_Data {
  public:
    typedef GT LensingType;
    typedef Lensing::LensingLosContainer LosLensingType;
    typedef typename boost::multi_array<LensingType, 1> LensingArray;
    typedef typename boost::multi_array<LosLensingType, 1>
        ProjectionLensingArray;

  protected:
    LensingArray Lensing;
    ProjectionLensingArray projection;
    long numObs;
    long numLos;

  public:
    LensingSurvey() : numObs(0), numLos(0) {}
    ~LensingSurvey() {}

    long NumberObs() const { return numObs; }
    long NumberLos() const { return numLos; }

    // Methods defined in the tcc file
    void addObs(const LensingType &lens);
    void addLos(LosLensingType &los);

    LensingArray &getLensing() { return Lensing; }
    const LensingArray &getLensing() const { return Lensing; }

    ProjectionLensingArray &getProjection() { return projection; }
    const ProjectionLensingArray &getProjection() const { return projection; }

    void optimize() { Lensing.resize(boost::extents[numObs]); }

    // I/O support for galaxy surveys
    void saveMain(H5::H5Location &fg);
    void restoreMain(H5::H5Location &fg);

    // Elenis_modification
    // void save(H5::H5Location &fg) { saveMain(fg); }
    void save(H5::H5Location &fg, bool partialSave = true) { saveMain(fg); }
    void restore(H5::H5Location &fg) { restoreMain(fg); }
  };

}; // namespace LibLSS

#  include "lensing_survey.tcc"

#endif
// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Natalia Porqueres
// ARES TAG: email(0) = n.porqueres@imperial.ac.uk
// ARES TAG: year(0) = 2020
