/*+
    ARES/HADES/BORG Package -- ./libLSS/data/lensing_survey.tcc
    Copyright (C) 2020-2021 Natalia Porqueres <n.porqueres@imperial.ac.uk>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021)

+*/
#include <iostream>
#include <CosmoTool/hdf5_array.hpp>
#include "libLSS/tools/hdf5_scalar.hpp"

namespace LibLSS {

  template <class LensingType, class AllocationPolicy>
  void LensingSurvey<LensingType, AllocationPolicy>::addObs(
      const LensingType &lens) {
    if (numObs == Lensing.size()) {
      Lensing.resize(boost::extents[numObs + AllocationPolicy::getIncrement()]);
    }

    Lensing[numObs] = lens;

    numObs++;
  }

  template <class LensingType, class AllocationPolicy>
  void
  LensingSurvey<LensingType, AllocationPolicy>::addLos(LosLensingType &los) {
    if (numLos == projection.size()) {
      projection.resize(
          boost::extents[numLos + AllocationPolicy::getIncrement()]);
    }

    projection[numLos] = los;
    numLos++;
  }

  template <class LensingType, class AllocationPolicy>
  void
  LensingSurvey<LensingType, AllocationPolicy>::saveMain(H5::H5Location &fg) {
    optimize();

    H5::Group g0 = fg.createGroup("lensing_data");
    CosmoTool::hdf5_write_array(g0, "lensing", Lensing);

    for (int i = 0; i < numObs; i++) {
      H5::Group g = g0.createGroup(str(boost::format("los_%d") % i));
      CosmoTool::hdf5_write_array(g, "voxel_id", projection[i].voxel_id);
      CosmoTool::hdf5_write_array(g, "dlos", projection[i].dlos);
      CosmoTool::hdf5_write_array(g, "rcoord", projection[i].rcoord);
      CosmoTool::hdf5_write_array(g, "redshift", projection[i].redshift);

      CosmoTool::hdf5_write_array(g, "ret00", projection[i].ret00);
      CosmoTool::hdf5_write_array(g, "ret11", projection[i].ret11);
      CosmoTool::hdf5_write_array(g, "ret22", projection[i].ret22);
      CosmoTool::hdf5_write_array(g, "ret33", projection[i].ret33);
      CosmoTool::hdf5_write_array(g, "ret_dev", projection[i].ret_dev);
      CosmoTool::hdf5_write_array(g, "iu0_v", projection[i].iu0_v);
      CosmoTool::hdf5_write_array(g, "weights", projection[i].weights);

      CosmoTool::hdf5_write_array(g, "alpha", projection[i].alpha);
    }
  }

  template <class LensingType, class AllocationPolicy>
  void LensingSurvey<LensingType, AllocationPolicy>::restoreMain(
      H5::H5Location &fg) {
    H5::Group g0 = fg.openGroup("lensing_data");
    CosmoTool::hdf5_read_array(g0, "lensing", Lensing);

    numObs = Lensing.size();

    projection.resize(boost::extents[numObs]);

    for (int i = 0; i < numObs; i++) {
      int s = 0;
      H5::Group g = g0.openGroup(str(boost::format("los_%d") % i));

      CosmoTool::hdf5_read_array(g, "voxel_id", projection[i].voxel_id);
      CosmoTool::hdf5_read_array(g, "dlos", projection[i].dlos);
      CosmoTool::hdf5_read_array(g, "rcoord", projection[i].rcoord);
      CosmoTool::hdf5_read_array(g, "redshift", projection[i].redshift);

      CosmoTool::hdf5_read_array(g, "ret00", projection[i].ret00);
      CosmoTool::hdf5_read_array(g, "ret11", projection[i].ret11);
      CosmoTool::hdf5_read_array(g, "ret22", projection[i].ret22);
      CosmoTool::hdf5_read_array(g, "ret33", projection[i].ret33);
      CosmoTool::hdf5_read_array(g, "ret_dev", projection[i].ret_dev);
      CosmoTool::hdf5_read_array(g, "iu0_v", projection[i].iu0_v);
      CosmoTool::hdf5_read_array(g, "weights", projection[i].weights);

      CosmoTool::hdf5_read_array(g, "alpha", projection[i].alpha);
    }
  }

}; // namespace LibLSS

// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Natalia Porqueres
// ARES TAG: email(0) = n.porqueres@imperial.ac.uk
// ARES TAG: year(0) = 2020-2021
