/*+
    ARES/HADES/BORG Package -- ./libLSS/data/schechter_completeness.hpp
    Copyright (C) 2014-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2022 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2015-2021)
       Minh Nguyen <minh@mpa-garching.mpg.de> (2017)

+*/
#ifndef __LIBLSS_SAMPLE_SELECTION_HPP
#  define __LIBLSS_SAMPLE_SELECTION_HPP

#  include <cmath>
#  include <functional>
#  include "libLSS/tools/console.hpp"
#  include <CosmoTool/hdf5_array.hpp>
#  include "libLSS/data/galaxies.hpp"
#  include "libLSS/data/projection.hpp"
#  include "libLSS/samplers/core/random_number.hpp"

namespace LibLSS {

  typedef std::function<bool(const BaseGalaxyDescriptor &)> GalaxySelector;
  typedef std::function<void(RandomNumber &, BaseGalaxyDescriptor &)>
      GalaxyGenerator;

  struct GalaxySelectionExtra {
    std::shared_ptr<GalaxySelector> selector;
    std::shared_ptr<GalaxyGenerator> generator;
  };

  struct GalaxySampleSelection {
    double bright_apparent_magnitude_cut;
    double faint_apparent_magnitude_cut;
    double bright_absolute_magnitude_cut;
    double faint_absolute_magnitude_cut;

    double zmin, zmax;
    double dmin, dmax;

    double low_mass_cut;
    double high_mass_cut;
    double small_radius_cut;
    double large_radius_cut;
    double low_spin_cut;
    double high_spin_cut;

    ProjectionDataModel projection;
  };

  static inline std::shared_ptr<GalaxySelector> makeSelector(GalaxySelector f) {
    return std::make_shared<GalaxySelector>(f);
  }

} // namespace LibLSS

// clang-format off
CTOOL_STRUCT_TYPE(
    LibLSS::GalaxySampleSelection, HDF5T_GalaxySampleSelection,
    ((double, bright_apparent_magnitude_cut))
    ((double, faint_apparent_magnitude_cut))
    ((double, bright_absolute_magnitude_cut))
    ((double, faint_absolute_magnitude_cut))
    ((double, zmin))
    ((double, zmax))
    ((double, dmin))
    ((double, dmax))
    ((double, low_mass_cut))
    ((double, high_mass_cut))
    ((double, small_radius_cut))
    ((double, large_radius_cut))
    ((double, low_spin_cut))
    ((double, high_spin_cut))
    ((LibLSS::ProjectionDataModel, projection))
);
// clang-format on

#endif
// ARES TAG: num_authors = 2
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2014-2024
// ARES TAG: email(1) = jens.jasche@fysik.su.se
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: year(1) = 2009-2022
