/*+
    ARES/HADES/BORG Package -- libLSS/physics/data/mixer_representation.hpp
    Copyright (C) 2021-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2022)

+*/
#pragma once
#ifndef __LIBLSS_PHYSICS_MIXER_DATA_REPRESENTATION_HPP
#  define __LIBLSS_PHYSICS_MIXER_DATA_REPRESENTATION_HPP

#  include "libLSS/physics/data/data_representation.hpp"

namespace LibLSS {
  namespace DataRepresentation {
    class MixerDataRepresentation : public AbstractRepresentation {
    public:
      typedef std::vector<AbstractRepresentation_p> SubRepresentations;

      MixerDataRepresentation(SubRepresentations &representations);
      MixerDataRepresentation(
          SubRepresentations &representations,
          AbstractRepresentation_p &&original, BackwardConverter cvt);

      ~MixerDataRepresentation() override;

      unsigned int dimensions() override;
      std::type_index getType() const override;

      virtual AbstractRepresentation_p shallowMove() override;

      virtual AbstractRepresentation_p close() override;

      static Descriptor
      make_descriptor(std::vector<Descriptor> const &desc_set);

      SubRepresentations &getSubs() { return representations; }

    protected:
      AbstractRepresentation_p base_representation;
      SubRepresentations representations;
      BackwardConverter closing_converter;
    };

  } // namespace DataRepresentation

} // namespace LibLSS

#endif
// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2021-2022
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
