/*+
    ARES/HADES/BORG Package -- libLSS/physics/data/tiledarray_representation.hpp
    Copyright (C) 2021 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021-2022)

+*/
#pragma once
#ifndef __LIBLSS_PHYSICS_DATA_TILEDARRAY_REPRESENTATION_HPP
#  define __LIBLSS_PHYSICS_DATA_TILEDARRAY_REPRESENTATION_HPP

#  include <map>
#  include <typeinfo>
#  include <functional>
#  include <typeindex>
#  include <memory>
#  include "libLSS/physics/model_io.hpp"
#  include "libLSS/physics/tiled_array.hpp"
#  include "libLSS/physics/data/data_representation.hpp"

namespace LibLSS {
  namespace DataRepresentation {

    template <typename T, size_t Nd>
    class TiledArrayRepresentation : public AbstractRepresentation {
    public:
      typedef TiledArrayRepresentation<T, Nd> self_t;
      typedef TiledArray<T, Nd> BaseArray;

    protected:
      std::shared_ptr<BaseArray> base;
      AbstractRepresentation_p base_representation;
      BackwardConverter closing_converter;

      TiledArrayRepresentation(TiledArrayRepresentation const &base_template);

    public:
      TiledArrayRepresentation() = delete;
      TiledArrayRepresentation(
          std::shared_ptr<BaseArray> a,
          std::unique_ptr<AbstractRepresentation> &&base_ptr =
              std::unique_ptr<AbstractRepresentation>(),
          BackwardConverter cvt = BackwardConverter());
      TiledArrayRepresentation(
          BaseArray &&a,
          std::unique_ptr<AbstractRepresentation> &&base_ptr =
              std::unique_ptr<AbstractRepresentation>(),
          BackwardConverter cvt = BackwardConverter());
      virtual ~TiledArrayRepresentation();

      std::type_index getType() const override { return typeid(self_t); }
      unsigned int dimensions() override { return Nd; }
      AbstractRepresentation_p shallowMove() override;
      AbstractRepresentation_p close() override;
      BaseArray &getContent();
      std::shared_ptr<BaseArray> getContentPtr();

      static Descriptor make_descriptor(
          std::shared_ptr<MPI_Communication> comm, std::array<size_t, Nd> dims);
      static void decode_descriptor(
          Descriptor &desc, std::shared_ptr<MPI_Communication> &comm,
          std::array<size_t, Nd> &dims);
    };

    extern template class TiledArrayRepresentation<double, 1ul>;
    extern template class TiledArrayRepresentation<double, 2ul>;
    extern template class TiledArrayRepresentation<double, 3ul>;

  } // namespace DataRepresentation
} // namespace LibLSS

#endif
// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2021
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
