/*+
    ARES/HADES/BORG Package -- libLSS/physics/data/tiledarray_representation_internal.hpp
    Copyright (C) 2021 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2022)

+*/
#pragma once
#ifndef __LIBLSS_PHYSICS_DATA_TILEDARRAY_REPRESENTATION_INTERNAL_HPP
#  define __LIBLSS_PHYSICS_DATA_TILEDARRAY_REPRESENTATION_INTERNAL_HPP
#  include <typeinfo>
#  include <functional>
#  include <typeindex>
#  include "libLSS/physics/model_io.hpp"
#  include "libLSS/physics/data/data_representation.hpp"
#  include "libLSS/physics/data/tiledarray_representation.hpp"
#  include "libLSS/tools/overload.hpp"
#  include "libLSS/tools/string_tools.hpp"
#  include "libLSS/tools/demangle.hpp"

namespace LibLSS {
  namespace DataRepresentation {
    template <typename T, size_t Nd>
    struct OpaqueTiledArrayDescriptor : OpaqueDescriptor {
      std::shared_ptr<MPI_Communication> comm;
      std::array<size_t, Nd> dims;

      std::shared_ptr<OpaqueDescriptor> clone() const override {
        auto new_desc = std::make_shared<OpaqueTiledArrayDescriptor>();

        new_desc->comm = comm;
        new_desc->dims = dims;

        return new_desc;
      }

      AbstractRepresentation_p makeTemporaryForward() const override {
        return std::make_unique<TiledArrayRepresentation<T, Nd>>(
            TiledArray<T, Nd>(comm, dims));
      }

      AbstractRepresentation_p makeTemporaryAdjointGradient() const override {
        return std::make_unique<TiledArrayRepresentation<T, Nd>>(
            TiledArray<T, Nd>(comm, dims));
      }
    };

  } // namespace DataRepresentation
} // namespace LibLSS
#endif

// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2021
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
