/*+
+*/
#pragma once
#include "libLSS/tools/compiler_tools.hpp"
#include "libLSS/physics/forward_model.hpp"

namespace LibLSS {

  // This class create a model that wraps a shared_ptr and
  // forward all calls to the underlying model.
  // This is for memory allocation insulation between the
  // two models.

  // The two classes mirror respectively the basic v3 API (ForwardModel)
  // and the v2 API (BORGForwardModel). The v3 API of BORGForwardModel is not
  // mirrored and we rely on the existing default adaptation function
  // to remap the two API

  class LIBLSS_EXPORT ForwardModelCapsule : public ForwardModel {
  protected:
    // TODO: make it weak ?
    std::shared_ptr<ForwardModel> parent_model;

    void updateCosmo() override;

  public:
    ForwardModelCapsule(std::shared_ptr<ForwardModel> const &_parent_model);
    virtual ~ForwardModelCapsule();

    void setName(std::string const &name) override;

    void forwardModel_v3(GeneralIO::Input input) override;
    GeneralIO::Output getResultForward_v3(GeneralIO::Output output) override;
    void adjointModel_v3(GeneralIO::InputAdjoint ag_input) override;
    GeneralIO::OutputAdjoint
    getResultAdjointGradient_v3(GeneralIO::OutputAdjoint ag_output) override;

    CosmologicalParameters getCosmology() override;
    void setModelParams(ModelDictionnary const &params) override;
    boost::any getModelParam(
        std::string const &name, std::string const &parameterName) override;
    void setAdjointRequired(bool on) override;
    void clearAdjointGradient() override;
    DataRepresentation::Descriptor getInputDescription() const override;
    DataRepresentation::Descriptor getOutputDescription() const override;
    void accumulateAdjoint(bool do_accumulate) override;

    std::unique_ptr<ForwardModel> clone() const override;
  };

  class LIBLSS_EXPORT BORGForwardModelCapsule : public BORGForwardModel {
  protected:
    std::shared_ptr<BORGForwardModel> parent_model;

  public:
    BORGForwardModelCapsule(
        std::shared_ptr<BORGForwardModel> const &_parent_model);
    virtual ~BORGForwardModelCapsule();

    void setName(std::string const &name) override;

    PreferredIO getPreferredInput() const override;
    PreferredIO getPreferredOutput() const override;

    void holdParticles() override;
    void forwardModelRsdField(ArrayRef &deltaf, double *vobs_ext) override;
    void forwardModel_v2(ModelInput<3> delta_init) override;
    void getDensityFinal(ModelOutput<3> delta_output) override;
    void adjointModel_v2(ModelInputAdjoint<3> ag_delta_input) override;
    void getAdjointModelOutput(ModelOutputAdjoint<3> ag_delta_output) override;

    void releaseParticles() override;
    bool densityInvalidated() const override;
  };

} // namespace LibLSS
