/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/fnl.hpp
    Copyright (C) 2020-2021 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2020 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2020-2021)

+*/
#ifndef __LIBLSS_HADES_FORWARD_FNL_HPP
#  define __LIBLSS_HADES_FORWARD_FNL_HPP
#  pragma once

#  include "libLSS/physics/forward_model.hpp"
#  include "libLSS/tools/fusewrapper.hpp"
#  include "libLSS/tools/errors.hpp"
#  include "libLSS/physics/forwards/registry.hpp"

namespace LibLSS {

  /**
   * This class handles the addition of an fNL into primordial potential.
   */
  class ForwardFNL : public BORGForwardModel {
  public:
    typedef IArrayType::ArrayType IArrayRef;
    typedef ArrayType1d::ArrayType Array1dRef;
    using BORGForwardModel::ArrayRef;

  protected:
    ModelInput<3> hold_input;
    ModelInputAdjoint<3> hold_ag_input;
    bool lazyInit;
    bool accumulateAg;
    double sign_fnl;

  public:
    /**
     * Constructor.
     */
    explicit ForwardFNL(
        std::shared_ptr<MPI_Communication> comm, const BoxModel &box);

    PreferredIO getPreferredInput() const override { return PREFERRED_REAL; }
    PreferredIO getPreferredOutput() const override { return PREFERRED_REAL; }

    void accumulateAdjoint(bool accumulate) override {
      accumulateAg = accumulate;
    }

    void forwardModel_v2(ModelInput<3> delta_init) override;
    void getDensityFinal(ModelOutput<3> delta_output) override;
    void adjointModel_v2(ModelInputAdjoint<3> in_gradient_delta) override;
    void
    getAdjointModelOutput(ModelOutputAdjoint<3> out_gradient_delta) override;
    void forwardModelRsdField(ArrayRef &, double *) override;
    void clearAdjointGradient() override;
    void releaseParticles() override;
    void changeSign();
    CosmologicalParameters getCosmoParams();
  }; // namespace LibLSS

} // namespace LibLSS

LIBLSS_REGISTER_FORWARD_DECL(PRIMORDIAL_FNL);

#endif
// ARES TAG: num_authors = 2
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2020-2021
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: year(1) = 2020
// ARES TAG: email(1) = jens.jasche@fysik.su.se
