/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/sphere_projection.hpp
    Copyright (C) 2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021)
       nataliaporqueres <natalia_porqueres@hotmail.com> (2021)

+*/
#pragma once
#ifndef __LIBLSS_PHYSICS_FORWARDS_SPHERE_PROJECTION_HPP
#  define __LIBLSS_PHYSICS_FORWARDS_SPHERE_PROJECTION_HPP

#  include "libLSS/physics/forward_model.hpp"
#  include "libLSS/tools/fusewrapper.hpp"
#  include "libLSS/tools/errors.hpp"
#  include "libLSS/physics/forwards/registry.hpp"
#  include "libLSS/physics/model_io/box.hpp"
#  include "libLSS/physics/data/data_representation.hpp"
#  include "libLSS/tools/auto_interpolator.hpp"

namespace LibLSS {

  /**
   * This class handles the convolution of a real field by some Fourier kernel.
   */
  class ForwardKappaSphere : public ForwardModel {
  public:
    DataRepresentation::AbstractRepresentation_p input_density_p,
        ag_input_density_p;
    BoxModel box;
    unsigned int Nside;

    typedef auto_interpolator<double> RedshiftInterpolator;

    double lensing_factor;
    RedshiftInterpolator redshift_interpolator;

    size_t start_Npix, local_Npix, end_Npix;
    std::array<ssize_t, 3> required_start;
    std::array<size_t, 3> required_shape, required_end;
    boost::multi_array<double, 2> source_distribution;

    void updateCosmo() override;

  public:
    /**
     * Constructor.
     */
    ForwardKappaSphere(
        std::shared_ptr<MPI_Communication> comm, const BoxModel &box_,
        unsigned int Nside_);

    virtual ~ForwardKappaSphere();

    void forwardModel_v3(GeneralIO::Input input) override;
    GeneralIO::Output getResultForward_v3(GeneralIO::Output output) override;
    void adjointModel_v3(GeneralIO::InputAdjoint input) override;
    GeneralIO::OutputAdjoint
    getResultAdjointGradient_v3(GeneralIO::OutputAdjoint input) override;

    void setModelParams(ModelDictionnary const &params) override;
    boost::any getModelParam(
        std::string const &model_name, std::string const &param) override;

    DataRepresentation::Descriptor getInputDescription() const override;
    DataRepresentation::Descriptor getOutputDescription() const override;
  };

} // namespace LibLSS

LIBLSS_REGISTER_FORWARDv3_DECL(KappaSphere);

#endif

// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2020
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
