/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/sum.hpp
    Copyright (C) 2018-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2018-2019 Doogesh Kodi Ramanah <ramanah@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2020-2021)

+*/
#pragma once
#ifndef __LIBLSS_MIXER_FORWARD_MODEL_HPP
#  define __LIBLSS_MIXER_FORWARD_MODEL_HPP

#  include <boost/multi_array.hpp>
#  include "libLSS/physics/cosmo.hpp"
#  include "libLSS/mpi/generic_mpi.hpp"
#  include "libLSS/physics/forward_model.hpp"
#  include "libLSS/physics/general_model_io.hpp"
#  include "libLSS/physics/data/data_representation.hpp"
#  include <vector>

namespace LibLSS {

  class MixerForwardModel : public ForwardModel {
  public:
    typedef std::vector<std::shared_ptr<ForwardModel>> ModelCollection;

    MixerForwardModel(
        std::shared_ptr<MPI_Communication> comm_,
        std::vector<DataRepresentation::Descriptor> const &inputs);

    ~MixerForwardModel() override;

    void forwardModel_v3(GeneralIO::Input input) override;

    GeneralIO::Output getResultForward_v3(GeneralIO::Output output) override;

    void adjointModel_v3(GeneralIO::InputAdjoint ag_input) override;
    GeneralIO::OutputAdjoint
    getResultAdjointGradient_v3(GeneralIO::OutputAdjoint ag_output) override;

    void setModelCollection(ModelCollection &models);
    void setActiveModel(unsigned int active);

    void setAdjointRequired(bool required) override;
    void clearAdjointGradient() override;

    void setModelParams(ModelDictionnary const &params) override;

    boost::any
    getModelParam(std::string const &name, std::string const &param) override;

    DataRepresentation::Descriptor getInputDescription() const override;

    DataRepresentation::Descriptor getOutputDescription() const override;

  protected:
    void updateCosmo() override;

    ModelCollection models;
    unsigned int currentlyActive;
    std::vector<DataRepresentation::AbstractRepresentation_p> inputs;
    std::vector<DataRepresentation::AbstractRepresentation_p> ag_inputs;
    std::vector<DataRepresentation::Descriptor> input_descriptors;
  };
}; // namespace LibLSS

#endif

// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2022
