#pragma once
#  include <boost/multi_array.hpp>
#  include <functional>
#  include "libLSS/mpi/generic_mpi.hpp"
#  include "libLSS/tools/mpi_fftw_helper.hpp"
#  include "libLSS/tools/uninitialized_type.hpp"
#  include "libLSS/tools/array_tools.hpp"
#  include "libLSS/physics/forwards/particle_balancer/sortinfo.hpp"
#include <optional>

namespace LibLSS {
  template <typename PartIdxArray>
  inline void initIndexes(PartIdxArray part_idx, size_t numParts) {
    typename PartIdxArray::index_gen i_gen;
    typedef typename PartIdxArray::index_range i_range;
    // initialize array with a range 0, 1, 2, 3, 4, ... numParts
    copy_array_rv(
        part_idx[i_gen[i_range(0, numParts)]],
        b_fused_idx<long, 1>(boost::lambda::_1));
  }

// This is an example of an integrated storage for all the ancillary parameters required
// by the particle redistribute algorithm
  struct BalanceInfo {
    typedef boost::multi_array<ssize_t, 1> IdxArray;

    IdxArray offsetReceive, offsetSend, numTransfer, numReceive;
    UninitializedArray<IdxArray> *u_idx = nullptr;
    size_t localNumParticlesBefore, localNumParticlesAfter;
    MPI_Communication *comm = nullptr;
    std::unique_ptr<SortInfo> sortInfo;
    std::unique_ptr<SortInfoDoubleMesh<>> twomeshsortInfo;

    void allocate(MPI_Communication *newComm, size_t partNum) {
      LIBLSS_AUTO_DEBUG_CONTEXT(ctx);
      auto extComm = boost::extents[newComm->size()];

      comm = newComm;
      clear();
      u_idx = new UninitializedArray<IdxArray>(boost::extents[partNum]);
      initIndexes(u_idx->get_array(), partNum);
      offsetReceive.resize(extComm);
      offsetSend.resize(extComm);
      numTransfer.resize(extComm);
      numReceive.resize(extComm);
    }

    void clear() {
      if (u_idx != 0) {
        delete u_idx;
        u_idx = 0;
      }
    }
  };
}
