/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/pm/steps/density_mp.hpp
    Copyright (C) 2014-2018 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2018 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021)

+*/
#pragma once
#ifndef __LIBLSS_PM_STEP_DENSITY_MP_HPP
#  define __LIBLSS_PM_STEP_DENSITY_MP_HPP

#  include "libLSS/physics/forwards/pm/steps/types.hpp"

namespace LibLSS {
  namespace PM {
    template <typename CIC>
    struct DensityBuilder_OMP {
      typedef BORGForwardModel::CArrayRef CArrayRef;

      PMGrid &shared_grid;
      bool mpiActive;
      bool initialSetupDone;
      GhostPlanes<double, 2> ghosts;

      inline DensityBuilder_OMP(PMGrid &shared_grid_)
          : shared_grid(shared_grid_),
            mpiActive(shared_grid.mgr->getComm()->size() > 1),
            initialSetupDone(false) {
        LIBLSS_AUTO_CONTEXT(LOG_DEBUG, ctx);
      }

      void setupGhosts(ssize_t localN2) {
        LIBLSS_AUTO_CONTEXT(LOG_DEBUG, ctx);
        auto &mgr = shared_grid.mgr;

        if (ghosts.dims()[1] == localN2)
          return;
        ctx.format(
            " Dimensions %dx%d != *x%d", ghosts.dims()[0], ghosts.dims()[1],
            localN2);

        if (!initialSetupDone) {
          std::array<ssize_t, 1> req_planes{
              (mgr->startN0 + mgr->localN0) % mgr->N0};
          std::vector<ssize_t> owned_planes(mgr->localN0);
          for (unsigned int i = 0; i < mgr->localN0; i++)
            owned_planes[i] = mgr->startN0 + i;

          initialSetupDone = true;
          ghosts.setup(
              mgr->getComm(), req_planes, owned_planes,
              std::array<ssize_t, 2>{mgr->N1, localN2}, mgr->N0);
        } else {
          ghosts.updatePlaneDims(std::array<ssize_t, 2>{mgr->N1, localN2});
        }
      }

      template <typename PositionArray, typename DensityArray>
      void forward(
          PositionArray &&positions, DensityArray &&density, double nmean,
          bool shift_by_one);

      template <typename PositionArray, typename DensityArray, typename Weight>
      void forward(
          PositionArray &&positions, DensityArray &&density, Weight &&weight,
          double nmean, bool shift_by_one);

      template <
          typename PositionArray, typename AgDensityArray,
          typename AgPositionArray>
      void adjoint(
          PositionArray &&positions, AgDensityArray &&ag_density,
          AgPositionArray &&ag_positions, double nmean);
    };
  } // namespace PM
} // namespace LibLSS

#  include "libLSS/physics/forwards/pm/steps/density_mp.tcc"
#endif
// ARES TAG: authors_num = 2
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2014-2018
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: email(1) = jens.jasche@fysik.su.se
// ARES TAG: year(1) = 2009-2018
