/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/pm/steps/density.hpp
    Copyright (C) 2014-2018 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2018 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2018-2021)

+*/
#pragma once
#ifndef __LIBLSS_PM_STEP_DENSITY_TILE_HPP
#  define __LIBLSS_PM_STEP_DENSITY_TILE_HPP

#  include <optional>
#  include "libLSS/physics/forward_model.hpp"
#  include "libLSS/physics/forwards/pm/steps/types.hpp"
#  include "libLSS/tools/xtensor_tools.hpp"
#  include "libLSS/physics/tiled_array.hpp"

namespace LibLSS {
  namespace PM {

    template <typename array>
    struct BoundingBox {
      typename array::IndexType start;
      typename array::DimType dims;
    };

    template <typename CIC>
    struct DensityTileBuilder {
      typedef BORGForwardModel::CArrayRef CArrayRef;

      PMGrid &shared_grid;
      bool mpiActive;
      bool initialSetupDone;
      typedef TiledArray<double, 3> array_t;
      typedef BoundingBox<array_t> BoundingBox_t;

      std::optional<BoundingBox_t> boundingbox;

      void clear_memory();

      inline DensityTileBuilder(PMGrid &shared_grid_)
          : shared_grid(shared_grid_),
            mpiActive(shared_grid.mgr->getComm()->size() > 1),
            initialSetupDone(false) {
        LIBLSS_AUTO_CONTEXT(LOG_DEBUG, ctx);
      }

      template <typename PositionArray>
      void getBoundingBox(
          PositionArray &&positions, array_t::IndexType &start,
          array_t::DimType &dims);

      template <typename PositionArray, typename DensityArray, typename Weight>
      void do_deposition(
          PositionArray const &positions, DensityArray &density,
          Weight const &weight);

      template <typename PositionArray>
      void forward(
          PositionArray &&positions, array_t &density, double nmean,
          bool shift_by_one);

      template <typename PositionArray, typename Weight>
      void forward(
          PositionArray &&positions, array_t &density, Weight &&weight,
          double nmean, bool shift_by_one);

      template <
          typename PositionArray, typename DensityArray,
          typename AgPositionArray, typename WeightArray>
      void adjoint_process(
          PositionArray const &positions, DensityArray const &ag_density,
          AgPositionArray &ag_positions, WeightArray const &weight,
          double nmean);

      template <typename PositionArray, typename AgPositionArray>
      void adjoint(
          PositionArray &&positions, array_t &ag_density,
          AgPositionArray &&ag_positions, double nmean);

      template <
          typename PositionArray, typename AgPositionArray,
          typename WeightArray>
      void adjoint(
          PositionArray &&positions, array_t &ag_density,
          AgPositionArray &&ag_positions, WeightArray &&weight, double nmean);
    };

    template <typename DensityArray>
    void convert_to_overdensity(
        DensityArray &density, double nmean, bool shift_by_one) {
      LIBLSS_AUTO_DEBUG_CONTEXT(ctx);
      const double inv_nmean = 1 / nmean;
      auto a = multi_array_to_xtensor(density);
      if (shift_by_one)
        xt::noalias(a) = a * inv_nmean - 1;
      else
        xt::noalias(a) = a * inv_nmean;
    }
  } // namespace PM
} // namespace LibLSS

#  include "libLSS/physics/forwards/pm/steps/density_tile.tcc"
#endif
// ARES TAG: authors_num = 2
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2014-2018
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: email(1) = jens.jasche@fysik.su.se
// ARES TAG: year(1) = 2009-2018
// ARES TAG: name(2) = Ewoud Wempe
// ARES TAG: email(2) = wempe@astro.rug.nl
// ARES TAG: year(2) = 2022
