/*+
+*/
#pragma once
#include "libLSS/physics/forwards/pm/steps/types.hpp"

namespace LibLSS::PM {
  template <
      typename Array1, typename Array2,
      typename E = typename std::remove_reference<Array1>::type::element>
  __attribute__((always_inline)) inline E kernel(
      Array1 &&a1, Array2 &&a2, E x, E y, E z, ssize_t ix, ssize_t iy,
      ssize_t iz, ssize_t jy, ssize_t jz) {

    E rx = (x - ix);
    E ry = (y - iy);
    E rz = (z - iz);

    E qx = 1 - rx;
    E qy = 1 - ry;
    E qz = 1 - rz;

    // clang-format off
      return
        a1[iy][iz] * qx * qy * qz +
        a1[iy][jz] * qx * qy * rz +
        a1[jy][iz] * qx * ry * qz +
        a1[jy][jz] * qx * ry * rz +
        a2[iy][iz] * rx * qy * qz +
        a2[iy][jz] * rx * qy * rz +
        a2[jy][iz] * rx * ry * qz +
        a2[jy][jz] * rx * ry * rz;
    // clang-format on
  }

  template <
      int axis, typename Array1, typename Array2,
      typename E = typename std::remove_reference<Array1>::type::element>
  __attribute__((always_inline)) inline E kernel_ag(
      Array1 &&a1, Array2 &&a2, E x, E y, E z, ssize_t ix, ssize_t iy,
      ssize_t iz, ssize_t jy, ssize_t jz) {

    static_assert(axis >= 0 && axis < 3, "Invalid axis specification.");

    E qx, qy, qz, rx, ry, rz;

    switch (axis) {
    case 0:
      rx = 1;
      ry = (y - iy);
      rz = (z - iz);
      qx = -1;
      qy = 1 - ry;
      qz = 1 - rz;
      break;
    case 1:
      rx = (x - ix);
      ry = 1;
      rz = (z - iz);
      qx = 1 - rx;
      qy = -1;
      qz = 1 - rz;
      break;
    case 2:
      rx = (x - ix);
      ry = (y - iy);
      rz = 1;
      qx = 1 - rx;
      qy = 1 - ry;
      qz = -1;
      break;
    }

    // clang-format off
      return
        a1[iy][iz] * qx * qy * qz +
        a1[iy][jz] * qx * qy * rz +
        a1[jy][iz] * qx * ry * qz +
        a1[jy][jz] * qx * ry * rz +
        a2[iy][iz] * rx * qy * qz +
        a2[iy][jz] * rx * qy * rz +
        a2[jy][iz] * rx * ry * qz +
        a2[jy][jz] * rx * ry * rz;
    // clang-format on
  }
} // namespace LibLSS::PM

// ARES TAG: num_authors = 2
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2016-2022
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: year(1) = 2012-2022
// ARES TAG: email(1) = j.jasche@fysik.su.se
