/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/pm/steps/ic.hpp
    Copyright (C) 2018 Guilhem lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2018 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2018, 2020)

+*/
#ifndef __LIBLSS_BORG_PM_STEPS_IC_HPP
#  define __LIBLSS_BORG_PM_STEPS_IC_HPP

#  include "libLSS/physics/forward_model.hpp"

namespace LibLSS {
  namespace PM {
    struct InitialConditions {
      typedef BORGForwardModel::CArrayRef CArrayRef;

      PMState &shared_obj;
      double last_vScale;
      double D0;
      double D1;
      double Df1;
      double f1;
      double Hubble; ///km /sec /(Mpc/h)
      double an, anh;
      bool do_Dunits;

      template <typename TimingArray>
      InitialConditions(
          PMState &shared_state, double ai, TimingArray &&timing,
          bool do_Dunits = false);

      template <
          typename PositionArray, typename VelocityArray, typename IndexArray>
      void forward(
          CArrayRef const &
              deltao, // The input complex array field is the only one that can be strict type.
          PositionArray &&
              pos, // Position and velocities may be an array or a  slice of array, or even other array-like types.
          VelocityArray &&vel,
          IndexArray &&part_idx // Generate particle ids
      );

      template <
          typename PositionArrayAG, typename VelocityArrayAG, typename AGPhase>
      void adjoint(
          PositionArrayAG &&pos_ag, VelocityArrayAG &&vel_ag,
          AGPhase &&phase_ag);
    };
  } // namespace PM
} // namespace LibLSS

#  include "libLSS/physics/forwards/pm/steps/ic.tcc"

#endif
// ARES TAG: authors_num = 2
// ARES TAG: name(0) = Guilhem lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2018
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: email(1) = jens.jasche@fysik.su.se
// ARES TAG: year(1) = 2018
