/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/pm/steps/particle_force.hpp
    Copyright (C) 2014-2019 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2018 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2018-2020)

+*/
#ifndef __LIBLSS_PM_PARTICLE_FORCE_HPP
#  define __LIBLSS_PM_PARTICLE_FORCE_HPP
#  include "libLSS/physics/forwards/pm/steps/force_kernel.hpp"

namespace LibLSS {
  namespace PM {

    /**
     * This helper class for PM provides the forward and
     * adjoint gradient operator for the force interpolation
     * on the mesh grid (a PM operation).
     */
    class ParticleForce {
    protected:
      PMState &shared_obj;
      PMGrid &grid;
      GhostPlanes<double, 2> ghosts;
      bool ghostEnabled, ghostInit;

      template <typename Array>
      void setupGhosts(Array &&array);

    public:
      /**
       * This constructors a new ParticleForce object
       * with the properties provided by the PMState object
       *
       * @param shared_state a PM state object.
       *
       * @see PMState
       */
      inline ParticleForce(PMState &shared_state, PMGrid &grid);
      inline ParticleForce(PMState &shared_state);

      /**
        * This is the forward operator.
        *
        * @param pos an array positions with shape [N][3], it may be a subarray
        * @param gravity an array holding the gravity field along some axis, must be a [n][m][p] array.
        * @param updater a functor that will be called for each particle with the interpolated force.
        * @see GravitySolver
        */
      template <
          typename PositionArray, typename GravityArray,
          typename ParticleUpdater>
      void forward(
          PositionArray &&pos, GravityArray &&gravity,
          ParticleUpdater &&updater);

      /**
        */
      template <
          typename PositionArray, typename GravityArray,
          typename ParticleUpdaterAg>
      void adjoint(
          PositionArray &&pos, GravityArray &&gravity,
          ParticleUpdaterAg &&particle_updater_ag);
      template <typename PositionArray, typename GravityArray>
      auto force_calculator(PositionArray &&pos, GravityArray &&gravity);
    };

  } // namespace PM
} // namespace LibLSS

#  include "libLSS/physics/forwards/pm/steps/particle_force.tcc"

#endif
// ARES TAG: authors_num = 2
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2014-2019
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: email(1) = jens.jasche@fysik.su.se
// ARES TAG: year(1) = 2009-2018
