/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/pm/steps/types.hpp
    Copyright (C) 2014-2019 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2018 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2018-2021)

+*/
#pragma once
#ifndef __LIBLSS_PM_STATE_HPP
#  define __LIBLSS_PM_STATE_HPP

#  include "libLSS/physics/forwards/pm/steps/types.hpp"
#  include "libLSS/physics/forwards/pm/steps/conv_grid.hpp"

namespace LibLSS::PM {

  struct PMState {
    std::shared_ptr<Cosmology> cosmo;
    ConvGrid<false> particles_resolution;
    ConvGrid<false> ic_resolution;
    ConvGrid<true> force_resolution;
    PMGrid output_resolution;
    //PMState(PMState const &other) = delete; // Just in case..

    //~PMState() { LIBLSS_AUTO_CONTEXT(LOG_DEBUG, ctx); }
  };

  inline BoxModel get_mpi_subbox(PMGrid const &grid) {
    // Gets a BoxModel of this MPI node's slab
    BoxModel out = grid;
    out.xmin0 = out.xmin0 + grid.mgr->startN0 * out.dx()[0];
    out.L0 = grid.mgr->localN0 * out.dx()[0];
    out.N0 = grid.mgr->localN0;
    return out;
  }

  template <bool do_COV>
  std::tuple<std::array<long, 3>, std::array<size_t, 3>>
  overlappingGrid_startend(
      BoxModel const &parent, BoxModel const &child,
      std::array<unsigned short, 6> padding = {0, 0, 0, 0, 0, 0}) {
    // Gets the (start, dims) of the tile that the child grid occupies in the parent grid.
    using namespace BORG_help;
    long ix_min = pos_to_idx<do_COV>(parent, child.xmin0, 0) - padding[0];
    long ix_max =
        pos_to_idx_ceil<do_COV>(parent, child.xmin0 + child.L0, 0) + padding[1];
    long iy_min = pos_to_idx<do_COV>(parent, child.xmin1, 1) - padding[2];
    long iy_max =
        pos_to_idx_ceil<do_COV>(parent, child.xmin1 + child.L1, 1) + padding[3];
    long iz_min = pos_to_idx<do_COV>(parent, child.xmin2, 2) - padding[4];
    long iz_max =
        pos_to_idx_ceil<do_COV>(parent, child.xmin2 + child.L2, 2) + padding[5];
    size_t diffx = ix_max - ix_min;
    size_t diffy = iy_max - iy_min;
    size_t diffz = iz_max - iz_min;
    assert(
        ix_max - ix_min >= 0 && iy_max - iy_min >= 0 && iz_max - iz_min >= 0);
    return {{ix_min, iy_min, iz_min}, {diffx, diffy, diffz}};
  }

} // namespace LibLSS::PM

#endif
// ARES TAG: authors_num = 3
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2014-2019
// ARES TAG: name(1) = Jens Jasche
// ARES TAG: email(1) = jens.jasche@fysik.su.se
// ARES TAG: year(1) = 2009-2018
// ARES TAG: name(2) = Ewoud Wempe
// ARES TAG: email(2) = wempe@astro.rug.nl
// ARES TAG: year(2) = 2022
