/*+
+*/
#pragma once
#include "libLSS/physics/forwards/pm/steps/types.hpp"
#include "libLSS/physics/tiled_array_tools.hpp"

namespace LibLSS::PM {
  struct DoubleMeshData {
    using array_t = TiledArray<double, 3>;
    using DimType = array_t::DimType;
    using IndexType = array_t::IndexType;
    using PaddingType = array_t::PaddingType;
    using array_view_t = TiledArrayView<double, 3, true>;
    array_t lr;
    array_t hr;
    array_view_t overlap_view;

    static array_t make_slabbed_tiledarray(
        std::shared_ptr<MPI_Communication> comm, PMGrid const &grid,
        PaddingType padding) {
      array_t arr{comm, grid.N};
      setupTiledSlabs(arr, *grid.mgr, padding);
      return arr;
    }
    DoubleMeshData(
        std::shared_ptr<MPI_Communication> comm, PMGrid const &grid,
        PMGrid const &zoom_grid, DimType dim_overlap, IndexType start_overlap,
        PaddingType padding)
        : lr{make_slabbed_tiledarray(comm, grid, padding)},
          hr(make_slabbed_tiledarray(comm, zoom_grid, padding)),
          overlap_view(lr, dim_overlap, start_overlap) {
      lr.useCoalescedIO();
      hr.useCoalescedIO();
    }
  };

  template <bool do_center_of_voxel = false>
  auto createDoubleMeshData(
      std::shared_ptr<MPI_Communication> comm, PMGrid const &grid,
      PMGrid const &zoom_grid, DoubleMeshData::PaddingType padding) {
    auto [start, dims] = overlappingGrid_startend<do_center_of_voxel>(
        grid, get_mpi_subbox(zoom_grid),
        {0, 1, 0, 1, 0, 1}); // Periodic BC don't help here!
    return DoubleMeshData(comm, grid, zoom_grid, dims, start, padding);
  }
} // namespace LibLSS::PM

// ARES TAG: authors_num = 1
// ARES TAG: name(2) = Ewoud Wempe
// ARES TAG: email(2) = wempe@astro.rug.nl
// ARES TAG: year(2) = 2022

