/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/forwards/primordial_as.hpp
    Copyright (C) 2021 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021)

+*/
#pragma once
#ifndef __LIBLSS_HADES_FORWARD_PRIMORDIAL_AS_HPP
#  define __LIBLSS_HADES_FORWARD_PRIMORDIAL_AS_HPP

#  include "libLSS/physics/forward_model.hpp"
#  include "libLSS/tools/fusewrapper.hpp"
#  include "libLSS/tools/errors.hpp"
#  include "libLSS/physics/forwards/registry.hpp"

namespace LibLSS {

  /**
   * This class handles the colouring of a white noise field with a cosmological
   * power spectrum. It is best to provide a fourier and use the returned fourier
   * field with the v2 interface. Otherwise extra FFT will be called to
   * to match to v1 interface.
   */
  class ForwardPrimordial_As : public BORGForwardModel {
  public:
    typedef IArrayType::ArrayType IArrayRef;
    typedef ArrayType1d::ArrayType Array1dRef;
    using BORGForwardModel::ArrayRef;

  protected:
    CosmologicalParameters old_cosmo_params;
    double af, D_init;
    LibLSS::multi_array<int, 3> powerSpectrumKeys;
    LibLSS::multi_array<double, 1> powerSpectrum;
    LibLSS::multi_array<double, 1> keyTranslate;

    ModelInput<3> hold_input;
    ModelInputAdjoint<3> hold_ag_input;
    bool lazyInit;
    bool invalid;
    bool accumulateAg;

    double A_s, n_s, k_pivot;

    void updatePower();

  public:
    /**
     * Constructor.
     */
    explicit ForwardPrimordial_As(
        std::shared_ptr<MPI_Communication> comm, const BoxModel &box);

    PreferredIO getPreferredInput() const override { return PREFERRED_FOURIER; }
    PreferredIO getPreferredOutput() const override {
      return PREFERRED_FOURIER;
    }

    auto getPowerSpectrumArray();

    void accumulateAdjoint(bool accumulate) override {
      accumulateAg = accumulate;
    }

    void forwardModel_v2(ModelInput<3> delta_init) override;

    void getDensityFinal(ModelOutput<3> delta_output) override;

    void updateCosmo() override;

    void adjointModel_v2(ModelInputAdjoint<3> in_gradient_delta) override;

    void clearAdjointGradient() override;

    void
    getAdjointModelOutput(ModelOutputAdjoint<3> out_gradient_delta) override;

    void forwardModelRsdField(ArrayRef &, double *) override;

    void releaseParticles() override;

    bool densityInvalidated() const override;

    void setModelParams(ModelDictionnary const &params) override;
  }; // namespace LibLSS

} // namespace LibLSS

LIBLSS_REGISTER_FORWARD_DECL(PRIMORDIAL_AS);

#endif
// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: year(0) = 2021
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
