/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/likelihoods/base.hpp
    Copyright (C) 2014-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2022 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2019-2021)

+*/
#pragma once
#ifndef __LIBLSS_PHYSICS_LIKELIHOOD_BASE_HPP
#  define __LIBLSS_PHYSICS_LIKELIHOOD_BASE_HPP

#  include <type_traits>
#  include <functional>
#  include <boost/tti/has_member_function.hpp>
#  include "libLSS/tools/static_auto.hpp"
#  include "libLSS/physics/likelihoods/info.hpp"
#  include "libLSS/tools/array_concepts.hpp"

namespace LibLSS {

  namespace Likelihood {

    namespace optional_features_details {

      template <
          typename L, typename... A,
          typename = decltype(std::declval<L>().delta_log_probability(
              std::declval<A>()...))>
      double delta_log_probability_impl(L &l, int, A &&...args) {
        return l.delta_log_probability(std::forward<A>(args)...);
      }

      template <
          typename L, typename Data, typename Field1, typename Field2,
          typename Mask>
      double delta_log_probability_impl(
          L &l, double, Data const &data, Field1 f1, Field2 f2,
          Mask const &mask) {
        return l.log_probability(data, f1, mask) -
               l.log_probability(data, f2, mask);
      }

    } // namespace optional_features_details

    template <
        typename L, typename Data, typename Field1, typename Field2,
        typename Mask>
    double delta_log_probability(
        L &l, Data const &data, Field1 f1, Field2 f2, Mask const &mask) {
      return optional_features_details::delta_log_probability_impl(
          l, 0, data, f1, f2, mask);
    }
  } // namespace Likelihood

} // namespace LibLSS
//
AUTO_REGISTRATOR_DECL(LikelihoodDoc);

#endif
