/*+
    ARES/HADES/BORG Package -- -- ./libLSS/physics/projector_lensing.hpp
    Copyright (C) 2014-2019 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2019 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Natalia Porqueres <natalia_porqueres@hotmail.com> (2018)
+*/

#ifndef __LIBLSS_PHYSICS_PROJECTOR_LENSING_HPP
#  define __LIBLSS_PHYSICS_PROJECTOR_LENSING_HPP

#  include <cmath>
#  include "boost/multi_array.hpp"
#  include "libLSS/data/lensing.hpp"
#  include "libLSS/physics/project_tool_lensing.hpp"

namespace LibLSS {
  namespace Lensing {
    typedef boost::multi_array<size_t, 2> VoxelIdType;
    typedef boost::multi_array<double, 2> RetType;
    typedef boost::multi_array<double, 1> LosType;

    struct LensingLosContainer {
      VoxelIdType voxel_id;
      LosType dlos;
      LosType rcoord;
      LosType redshift;

      LosType alpha;
      RetType ret00;
      RetType ret11;
      RetType ret22;
      RetType ret33;
      RetType ret_dev;
      VoxelIdType iu0_v;
      RetType weights;

      LensingLosContainer &operator=(LensingLosContainer const &other) {
        if (&other == this)
          return *this;

        voxel_id.resize(boost::extents[other.voxel_id.shape()[0]][3]);
        dlos.resize(boost::extents[other.dlos.shape()[0]]);
        rcoord.resize(boost::extents[other.rcoord.shape()[0]]);
        redshift.resize(boost::extents[other.redshift.shape()[0]]);

        alpha.resize(boost::extents[other.alpha.shape()[0]]);
        ret00.resize(boost::extents[other.ret00.shape()[0]][8]);
        ret11.resize(boost::extents[other.ret11.shape()[0]][8]);
        ret22.resize(boost::extents[other.ret22.shape()[0]][8]);
        ret33.resize(boost::extents[other.ret33.shape()[0]][8]);
        ret_dev.resize(boost::extents[other.ret_dev.shape()[0]][8]);

        iu0_v.resize(boost::extents[other.iu0_v.shape()[0]][3]);
        weights.resize(
            boost::extents[other.weights.shape()[0]][4]); //Ncat=4 hardcoded

        return *this;
      }
    };

    static inline void
    get_vox_id(double *position, size_t *idx, double *min, double *dd) {
      for (int i = 0; i < 3; i++)
        idx[i] = std::floor((position[i] - min[i]) / dd[i]);
    }

    static inline void
    get_vox_id(double *position, double *idx, double *min, double *dd) {
      for (int i = 0; i < 3; i++)
        idx[i] = (position[i] - min[i]) / dd[i];
    }

    static inline void
    get_coord(double *position, size_t *idx, double *min, double *dd) {
      for (int i = 0; i < 3; i++)
        position[i] = idx[i] * dd[i] + min[i];
    }

    static inline void
    resize_container(LensingLosContainer &data, int idx, int Ncat) {
      data.dlos.resize(boost::extents[idx + 1]);
      data.rcoord.resize(boost::extents[idx + 1]);
      data.redshift.resize(boost::extents[idx + 1]);
      data.voxel_id.resize(boost::extents[idx + 1][3]);

      data.alpha.resize(boost::extents[idx + 1]);
      data.ret00.resize(boost::extents[idx + 1][8]);
      data.ret11.resize(boost::extents[idx + 1][8]);
      data.ret22.resize(boost::extents[idx + 1][8]);
      data.ret33.resize(boost::extents[idx + 1][8]);
      data.ret_dev.resize(boost::extents[idx + 1][8]);

      data.iu0_v.resize(boost::extents[idx + 1][3]);
      data.weights.resize(boost::extents[idx + 1][Ncat]);
    }

    static inline void store_container(
        LensingLosContainer &data, int idx, double pos, double dlos,
        size_t *voxel, double *dl, double *corner) {
      //data.dlos[idx] = pos;
      for (int j = 0; j < 3; j++)
        data.voxel_id[idx][j] = voxel[j];

      double dist = 0.;
      for (int j = 0; j < 3; j++)
        dist += pow(voxel[j] * dl[j] + corner[j], 2);
      data.rcoord[idx] = pow(dist, 0.5);
    }

    double ray_tracer(
        ssize_t *observer, double glx_distance, double *a_u, double *corner, double *dl,
        size_t *N, LensingLosContainer &data, long Ncat);

 } // namespace Lensing
} // namespace LibLSS
#endif
// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Natalia Porqueres
// ARES TAG: email(0) = n.porqueres@imperial.ac.uk
// ARES TAG: year(0) = 2020-2021
// ARES TAG: name(1) = Guilhem Lavaux
// ARES TAG: email(1) = guilhem.lavaux@iap.fr
// ARES TAG: year(1) = 2014-2019
