/*+
+*/
#pragma once
#include <vector>
#include <oneapi/tbb.h>

namespace LibLSS::TBBCIC {
  using namespace oneapi::tbb;
  template <bool invert_subranges = false>
  static std::vector<size_t> make_sorting_mapper(
      size_t Np, std::vector<size_t> const &cumsum,
      std::vector<size_t> const &key_arr,
      std::vector<size_t> &&number_of_elements) {
    std::vector<size_t> mapper(Np);
    for (
        size_t i = 0; i < Np;
        i++) { // This one is hard to optimize. Only possibility would be domain splitting, but that is extra work
      size_t my_ix = key_arr[i];
      size_t my_new_index =
          invert_subranges ? cumsum[my_ix] + --number_of_elements[my_ix]
                           : cumsum[my_ix + 1] - number_of_elements[my_ix]--;
      mapper[i] = my_new_index;
    }
    return mapper;
  }
  template <typename ToSortArray, typename OutputArray>
  typename std::enable_if_t<ToSortArray::dimensionality == 2> do_sorting_par(
      ToSortArray const &particles, OutputArray &particles_sorted, size_t Np,
      std::vector<size_t> const &mapper) {
    static affinity_partitioner ap;
    parallel_for(
        blocked_range<size_t>(0, Np),
        [&](blocked_range<size_t> const &r) {
          for (
              size_t i = r.begin(); i != r.end();
              i++) { // Using a mutex, using a list of mutexes or building separately
            size_t my_new_index = mapper[i];
            for (size_t axis = 0; axis < 3; axis++)
              particles_sorted[my_new_index][axis] = particles[i][axis];
          }
        },
        ap);
  }
  template <typename ToSortArray>
  typename std::enable_if_t<ToSortArray::dimensionality == 1> do_sorting_par(
      ToSortArray const &array, ToSortArray &arraySorted, size_t Np,
      std::vector<size_t> const &mapper) {
    affinity_partitioner ap;
    parallel_for(
        blocked_range<size_t>(0, Np),
        [&](blocked_range<size_t> const &r) {
          for (
              size_t i = r.begin(); i != r.end();
              i++) { // Using a mutex, using a list of mutexes or building separately
            size_t my_new_index = mapper[i];
            arraySorted[my_new_index] = array[i];
          }
        },
        ap);
  }
}

// ARES TAG: num_authors = 1
// ARES TAG: name(0) = Ewoud Wempe
// ARES TAG: year(0) = 2022
// ARES TAG: email(0) = wempe@astro.rug.nl
