/*+
    ARES/HADES/BORG Package -- ./libLSS/physics/velocity/velocity_cic.hpp
    Copyright (C) 2019-2020 Florent Leclercq <florent.leclercq@polytechnique.org>
    Copyright (C) 2019-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Florent Leclercq <florent.leclercq@polytechnique.org> (2020-2021)
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2020-2021)

+*/
#ifndef __LIBLSS_PHYSICS_VELOCITY_CIC
#  define __LIBLSS_PHYSICS_VELOCITY_CIC

#  include <boost/multi_array.hpp>

#  include "libLSS/physics/velocity/velocity.hpp"
#  include "libLSS/physics/classic_cic.hpp"
#  include "libLSS/samplers/core/powerspec_tools.hpp"
#  include "libLSS/physics/forwards/particle_balancer/particle_distribute.hpp"

namespace LibLSS {

  namespace VelocityModel {

    class CICModel : public ParticleBasedModel {
    public:
      typedef ClassicCloudInCell<double> CIC;
      typedef boost::multi_array<double, 2> PhaseArray;
      typedef UninitializedArray<PhaseArray> U_PArray;

    protected:
      std::shared_ptr<LibLSS::FFTW_Manager<double, 3>> mgr;
      double rsmooth;
      double part_factor;
      BalanceInfo realInfo;
      std::shared_ptr<U_PArray> u_pos, u_vel;
      bool smoothing;

    public:
      CICModel(
          BoxModel box_out_, particleForwardModel_t model_, double rsmooth_,
          double part_factor_)
          : ParticleBasedModel(box_out_, model_),
            mgr(std::make_shared<FFTW_Manager<double, 3>>(
                box_out_.N0, box_out_.N1, box_out_.N2, model_->communicator())),
            rsmooth(rsmooth_), part_factor(part_factor_) {
        smoothing = rsmooth > 0.;
      }

      std::shared_ptr<LibLSS::FFTW_Manager<double, 3>> const &getMgr() const {
        return mgr;
      }

      void queryLocalExtents(std::array<ssize_t, 6> &e) override {
        e[0] = mgr->startN0;
        e[1] = mgr->startN0 + mgr->localN0;
        e[2] = 0;
        e[3] = mgr->N1;
        e[4] = 0;
        e[5] = mgr->N2;
      }

      template <typename A_t>
      void getVelocityFieldAlpha(
          A_t VelocityFieldAlpha, arrayScalarField_t MassField,
          const unsigned int alpha);
      void getVelocityField(arrayVelocityField_t VelocityField) override;
      void computeAdjointModel(arrayVelocityField_t AGVelocityField) override;

    }; // class CICModel
  }    // namespace VelocityModel
};     // namespace LibLSS

#endif

// ARES TAG: authors_num = 2
// ARES TAG: name(0) = Florent Leclercq
// ARES TAG: year(0) = 2019-2024
// ARES TAG: email(0) = florent.leclercq@iap.fr
// ARES TAG: name(1) = Guilhem Lavaux
// ARES TAG: year(1) = 2019-2024
// ARES TAG: email(1) = guilhem.lavaux@iap.fr
