/*+
    ARES/HADES/BORG Package -- ./libLSS/samplers/borg/borg_studentT_likelihood.hpp
    Copyright (C) 2014-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2022 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021)

+*/
#ifndef __LIBLSS_BORG_STUDENT_T_DENSITY_HPP
#define __LIBLSS_BORG_STUDENT_T_DENSITY_HPP

#include "libLSS/mpi/generic_mpi.hpp"
#include "libLSS/mcmc/global_state.hpp"
#include "libLSS/samplers/core/markov.hpp"
#include "libLSS/samplers/core/random_number.hpp"
#include "libLSS/samplers/core/types_samplers.hpp"
#include "libLSS/physics/forward_model.hpp"
#include "libLSS/samplers/hades/base_likelihood.hpp"
#include "libLSS/samplers/core/registry.hpp"
#include "libLSS/physics/likelihoods/base.hpp"

namespace LibLSS {

  class BorgStudentTLikelihood : public HadesBaseDensityLikelihood {

  protected:
    // Number of degrees of freedom for StudentT, which is initialized
    // from the 'LikelihoodInfo'
    double dof = 10;

  public:
    typedef HadesBaseDensityLikelihood super_t;
    typedef HadesBaseDensityLikelihood::super_t grid_t;

  public:
    BorgStudentTLikelihood(LikelihoodInfo &info);
    ~BorgStudentTLikelihood() {}

    void initializeLikelihood(MarkovState &state) override;
    void updateMetaParameters(MarkovState &state) override;
    void setupDefaultParameters(MarkovState &state, int catalog) override;

    void generateMockSpecific(
        ArrayRef const &parameters, MarkovState &state) override;

    double get_sample(RandomGen *rgen);
    double logLikelihoodSpecific(ArrayRef const &parameters) override;
    void gradientLikelihoodSpecific(
        ArrayRef const &parameters, ArrayRef &gradient_parameters) override;
  };

}; // namespace LibLSS

LIBLSS_REGISTER_LIKELIHOOD_DECL(BORG_STUDENT_T);

#endif
