/*+
    ARES/HADES/BORG Package -- ./libLSS/samplers/hades/hades_linear_likelihood.hpp
    Copyright (C) 2014-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2022 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2016, 2018-2021)

+*/
#ifndef __LIBLSS_HADES_LINEAR_DENSITY_HPP
#define __LIBLSS_HADES_LINEAR_DENSITY_HPP

#include <CosmoTool/fourier/fft/fftw_calls.hpp>
#include "libLSS/mpi/generic_mpi.hpp"
#include "libLSS/mcmc/global_state.hpp"
#include "libLSS/tools/fftw_allocator.hpp"
#include "libLSS/samplers/core/markov.hpp"
#include "libLSS/samplers/core/random_number.hpp"
#include "libLSS/samplers/core/types_samplers.hpp"
#include "libLSS/samplers/core/gridLikelihoodBase.hpp"
#include "libLSS/physics/forward_model.hpp"
#include "libLSS/samplers/hades/base_likelihood.hpp"
#include "libLSS/samplers/core/registry.hpp"

namespace LibLSS {

  class HadesLinearDensityLikelihood : public HadesBaseDensityLikelihood {
  public:
    typedef HadesBaseDensityLikelihood super_t;
    typedef HadesBaseDensityLikelihood::super_t grid_t;

  public:
    HadesLinearDensityLikelihood(LikelihoodInfo &info);
    virtual ~HadesLinearDensityLikelihood();

    void setupDefaultParameters(MarkovState &state, int catalog) override;
    void generateMockSpecific(
        ArrayRef const &parameters, MarkovState &state) override;
    double logLikelihoodSpecific(ArrayRef const &parameters) override;
    double
    diffLogLikelihoodSpecific(ArrayRef const &d1, ArrayRef const &d2) override;
    void gradientLikelihoodSpecific(
        ArrayRef const &parameters, ArrayRef &gradient_parameters) override;
    void initializeLikelihood(MarkovState &state) override;
  };

}; // namespace LibLSS

LIBLSS_REGISTER_LIKELIHOOD_DECL(HADES_LINEAR);

#endif
