/*+
    ARES/HADES/BORG Package -- ./libLSS/samplers/lensing/hades_lensing_likelihood.hpp
    Copyright (C) 2020 Natalia Porqueres <n.porqueres@imperial.ac.uk>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2021)

+*/
#ifndef __LIBLSS_BORG_LENSING_DENSITY_HPP
#define __LIBLSS_BORG_LENSING_DENSITY_HPP

#  include <CosmoTool/fourier/fft/fftw_calls.hpp>
#  include "libLSS/mpi/generic_mpi.hpp"
#  include "libLSS/mcmc/global_state.hpp"
#  include "libLSS/tools/fftw_allocator.hpp"
#  include "libLSS/samplers/core/markov.hpp"
#  include "libLSS/samplers/core/random_number.hpp"
#  include "libLSS/samplers/core/types_samplers.hpp"
#  include "libLSS/physics/forward_model.hpp"
#  include "libLSS/samplers/lensing/base_likelihood_lensing.hpp"
#  include "libLSS/physics/likelihoods/base.hpp"
#  include "libLSS/tools/fftw_allocator.hpp"

#  include "libLSS/samplers/rgen/hmc/hmc_density_sampler.hpp"
#  include "libLSS/data/lensing.hpp"
#  include "libLSS/data/lensing_survey.hpp"
#  include "libLSS/tools/mpi/ghost_planes.hpp"
#  include "libLSS/tools/mpi_fftw_helper.hpp"
#  include <CosmoTool/fourier/fft/fftw_calls.hpp>
#  include "libLSS/samplers/core/registry.hpp"


namespace LibLSS {

  class BorgLensingLikelihood : public HadesBaseDensityLensingLikelihood {
  public:
    typedef HadesBaseDensityLensingLikelihood super_t;
    typedef HadesBaseDensityLensingLikelihood::super_t grid_t;
    typedef U_Array<
        std::complex<double>, 2, FFTW_Allocator<std::complex<double>>>
        U_MyArray2d;
    typedef ArrayStateElement<double, 2, FFTW_Allocator<double>, true>
        ArrayType2d;
    //typedef U_Array<double, 2 , FFTW_Allocator<double>> U_MyRealArray2d;

  protected:
    MarkovState *state;

    typedef FFTW_Manager<double, 2> Mgr2d;
    std::shared_ptr<Mgr2d> mgr_2d;
    FFTW_Allocator<std::complex<double>> alloc_complex_2d;
    FFTW_Allocator<double> alloc_real_2d;

    double sigma[4];
    double mean[4];
    double stdev[4];
    double A1, bTA, A2;

    std::vector<double> rcoord;

    bool need_init_ghosts = 1;
    GhostPlanes<double, 2> ghosts;

    long numCosmoParams = 2;
    long Niterations = 0;

    ssize_t localN0_2d = 0, startN0_2d = 0;

    ArrayType2d *mask;

  private:
    double computeDistanceToObserver(int ix, int iy, int iz);
    double computeDistanceLos(int iz);
    double weights(double iz, double mean, double std);

  public:
    BorgLensingLikelihood(LikelihoodInfo &info);
    virtual ~BorgLensingLikelihood();

    void setupDefaultMask();

    virtual void updateMetaParameters(MarkovState &state);
    virtual void setupDefaultParameters(MarkovState &state, int catalog);
    virtual void initializeLikelihood(MarkovState &state);

    virtual void
    generateMockSpecific(ArrayRef const &parameters, MarkovState &state);

    virtual double logLikelihoodSpecific(ArrayRef const &parameters);
    virtual void gradientLikelihoodSpecific(
        ArrayRef const &parameters, ArrayRef &gradient_parameters);
    void initialize_ghosts();
    double Jacobian(CosmologicalParameters &cosmo_params, double z);
    void sampleMeta(MarkovState &state);
  };

}; // namespace LibLSS

LIBLSS_REGISTER_LIKELIHOOD_DECL(BORG_SHEAR);

#endif
// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Natalia Porqueres
// ARES TAG: email(0) = n.porqueres@imperial.ac.uk
// ARES TAG: year(0) = 2020-2021
