/*+
+*/
#pragma once
#include "libLSS/mpi/generic_mpi.hpp"
#include "libLSS/samplers/core/main_loop.hpp"
#include "libLSS/samplers/core/markov.hpp"
#include <list>

namespace LibLSS {

  struct WarmerOptions {
    double target_acceptance_min = 0.65;
    double target_acceptance_max = 0.80;
    double step_tuner = 0.80;
    long window_acceptance = 50;
    int min_window = 10;
    double absolute_max_epsilon = 0.02;
    double absolute_min_epsilon = 5e-4;
    double epsilon_change_age = 10;
    double x_hat_warm = 0.7;
    int max_warmup = -1;
    bool ignore_xhat = false;
    bool always_tweaking = false;
    double target_acceptance = 0.66;
  };

  class WarmerSampler : public MarkovSampler {
  protected:
    MainLoop &loop;
    WarmerOptions opts;
    std::shared_ptr<MPI_Communication> comm;
    std::list<std::tuple<long, long>> history;

    void adjust_loop(std::list<StateTuple> tuples);

  public:
    WarmerSampler(
        std::shared_ptr<MPI_Communication> comm, MainLoop &loop,
        WarmerOptions const &opts);
    ~WarmerSampler();

    void initialize(MarkovState &state) override;
    void restore(MarkovState &state) override;
    void sample(MarkovState &state) override;

    void sample(std::list<StateTuple>) override;
    void initialize(std::list<StateTuple>) override;
    void restore(std::list<StateTuple>) override;
  };
} // namespace LibLSS

// ARES TAG: num_authors = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2024
