/*+
+*/
#pragma once
#include <utility>

namespace LibLSS {
  template <typename Base, typename Attr>
  struct AttributeExtensionVal : public Base {
    Attr extra;
    AttributeExtensionVal(Base &&base, Attr &&attr)
        : Base(std::move(base)), extra(std::move(attr)) {}
    AttributeExtensionVal(Base &base, Attr &&attr)
        : Base(base), extra(std::move(attr)) {}
  };

  template <typename Base, typename Attr>
  struct AttributeExtensionRef : public Base {
    Attr &extra;
    AttributeExtensionRef(Base &&base, Attr &attr)
        : Base(std::move(base)), extra(attr) {}
    AttributeExtensionRef(Base &base, Attr &attr) : Base(base), extra(attr) {}
  };

  template <bool is_needed, typename Base, typename Attr>
  auto appendOptionalAttribute(Base &&base, Attr &&attr) {
    if constexpr (is_needed)
      return appendAttribute(
          std::forward<Base>(base), std::forward<Attr>(attr));
    else
      return std::forward<Base>(base);
  }

  template <typename Base, typename Attr>
  auto appendAttribute(Base &&base, Attr &attr) {
    return AttributeExtensionRef<std::remove_reference_t<Base>, Attr>(
        std::forward<Base>(base), attr);
  }

  template <typename Base, typename Attr>
  auto appendAttribute(Base &&base, Attr &&attr) {
    return AttributeExtensionVal<std::remove_reference_t<Base>, Attr>(
        std::forward<Base>(base), std::move(attr));
  }

  MEMBER_VAR_CHECKER(hasAttribute_base, extra);

  template <typename T>
  using hasAttribute = hasAttribute_base<T, void>;

} // namespace LibLSS

// ARES TAG: num_authors = 1
// ARES TAG: name(0) = Ewoud Wempe
// ARES TAG: year(0) = 2022
// ARES TAG: email(0) = wempe@astro.rug.nl
