#pragma once

namespace LibLSS {
  inline namespace domain_utils {

    template <size_t I, size_t N>
    struct _slice_maker {
      template <typename E>
      static auto
      make(E e, typename DomainSpec<N>::DomainLimit_t const &slice) {
        typedef boost::multi_array_types::index_range r;
        return _slice_maker<I + 1, N>::make(
            e[r(slice[2 * I], slice[2 * I + 1])], slice);
      }

      template <typename E>
      static auto make(
          E e, typename DomainSpec<N>::DomainLimit_t const &slice,
          typename DomainSpec<N>::DomainShift_t const &shifter) {
        typedef boost::multi_array_types::index_range r;
        return _slice_maker<I + 1, N>::make(
            e[r(slice[2 * I] + shifter[I], slice[2 * I + 1] + shifter[I])],
            slice, shifter);
      }
    };

    template <size_t N>
    struct _slice_maker<N, N> {
      template <typename E>
      static inline auto
      make(E e, typename DomainSpec<N>::DomainLimit_t const &slice) {
        return e;
      }
      template <typename E>
      static inline auto make(
          E e, typename DomainSpec<N>::DomainLimit_t const &slice,
          typename DomainSpec<N>::DomainShift_t const &shifter) {
        return e;
      }
    };

    template <size_t N>
    auto _make_slice(typename DomainSpec<N>::DomainLimit_t const &slice) {
      return _slice_maker<0, N>::make(boost::indices, slice);
    }

    template <size_t N>
    auto _make_slice(
        typename DomainSpec<N>::DomainLimit_t const &slice,
        typename DomainSpec<N>::DomainShift_t const &shifter) {
      return _slice_maker<0, N>::make(boost::indices, slice, shifter);
    }

    template <size_t N>
    ssize_t
    computeSliceSize(typename DomainSpec<N>::DomainLimit_t const &slice) {
      ssize_t ret = 1;
      for (unsigned int i = 0; i < N; i++) {
        ret *= slice[2 * i + 1] - slice[2 * i];
      }
      return ret;
    }

    template <typename T, size_t N, typename Creator>
    auto makeTempSlice(
        typename DomainSpec<N>::DomainLimit_t const &slice, Creator create) {
      std::array<ssize_t, N> bases, shapes;

      for (unsigned int i = 0; i < N; i++) {
        bases[i] = slice[2 * i];
        shapes[i] = slice[2 * i + 1] - slice[2 * i];
      }
      return create(array::make_extent<N>(bases.data(), shapes.data()));
    }

    template <typename T, size_t N>
    auto makeTempSlice(typename DomainSpec<N>::DomainLimit_t const &slice) {
      return makeTempSlice<T, N>(slice, [](auto ext) {
        return std::make_shared<TemporarySlice<T, N>>(ext);
      });
    }
  } // namespace domain_utils
} // namespace LibLSS