/*+
    ARES/HADES/BORG Package -- libLSS/tools/domains_coalesced.hpp
    Copyright (C) 2023 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)

+*/
#pragma once
#ifndef __LIBLSS_TOOLS_DOMAINS_COALESCED_HPP
#  define __LIBLSS_TOOLS_DOMAINS_COALESCED_HPP

#  include <cstdint>
#  include <variant>
#  include "libLSS/tools/domains.hpp"

namespace LibLSS {
  template <size_t N>
  struct DomainTaskSpec {
    typename DomainSpec<N>::DomainLimit_t slice;
    typename DomainSpec<N>::DomainShift_t shifter;
    SliceOperation operation;
    size_t global_offset, size;
  };

  template <size_t N>
  struct DomainTaskCoalesced {
    std::list<DomainTaskSpec<N>> task_iters;
    int order;
    int rankIO;
    bool recv;
  };

  template <typename A>
  struct SliceWrap {
    A object;
    SliceWrap(A &&a) : object(std::move(a)) {}
    SliceWrap(A const &a) : object(a) {}
  };

  template <typename T, size_t N>
  using TemporarySliceCoalesced =
      std::variant<TemporarySlice<T, N>, SliceWrap<TemporarySlice<T, 1>>>;

  template <size_t N_>
  struct DomainTodoCoalesced {
    static constexpr size_t const &N = N_;
    std::list<std::variant<DomainTaskCoalesced<N_>, DomainTask<N_>>> tasks;
  };

  template <size_t N>
  void mpiDomainCoalesceTodo(
      DomainTodo<N> const &todo, DomainTodoCoalesced<N> &todo_coaelesced);
  template <size_t N>
  DomainTodoCoalesced<N> mpiDomainCoalesceTodo(DomainTodo<N> const &todo);

  template <typename T, size_t N>
  void mpiDomainRun(
      MPI_Communication *comm, Domain<T, N> const &input_domains,
      Domain<T, N> &output_domains, DomainTodoCoalesced<N> const &todo);

} // namespace LibLSS

#endif

// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2023
