/*+
    ARES/HADES/BORG Package -- ./libLSS/tools/openmp.hpp
    Copyright (C) 2014-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2022 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Florent Leclercq <florent.leclercq@polytechnique.org> (2021)
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2015-2016, 2018, 2020)

+*/
#ifndef __LIBLSS_OPENMP_HPP
#define __LIBLSS_OPENMP_HPP

#ifdef _OPENMP
#  include <omp.h>
#endif

namespace LibLSS {

  inline int smp_get_max_threads() {
#ifdef _OPENMP
    return omp_get_max_threads();
#else
    return 1;
#endif
  }

  inline int smp_get_thread_id() {
#ifdef _OPENMP
    return omp_get_thread_num();
#else
    return 0;
#endif
  }

  inline int smp_get_num_threads() {
#ifdef _OPENMP
    return omp_get_num_threads();
#else
    return 1;
#endif
  }

  inline void smp_set_num_threads(int n) {
#ifdef _OPENMP
    omp_set_num_threads(n);
#endif
  }

  inline void smp_set_nested(bool n) {
#ifdef _OPENMP
    omp_set_nested(n ? 1 : 0);
#endif
  }

  // This is hack to check whether openmp is running.
  // However, we should go away from openmp as soon as possible.
  inline bool smp_is_active() { return smp_get_num_threads() > 1; }

}; // namespace LibLSS

#endif
