/*+
    ARES/HADES/BORG Package -- ./libLSS/tools/ptree_vectors.hpp
    Copyright (C) 2014-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2022 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Florent Leclercq <florent.leclercq@polytechnique.org> (2021)
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2016-2020)

+*/
#ifndef __LIBLSS_PTREE_VECTORS_HPP
#define __LIBLSS_PTREE_VECTORS_HPP

#include <string>
#include <boost/lexical_cast.hpp>
#include <vector>
#include "libLSS/tools/string_tools.hpp"
#include <boost/exception/diagnostic_information.hpp>

namespace LibLSS {

  template <typename T>
  std::vector<T>
  string_as_vector(const std::string &s, std::string const &separator) {
    auto tokens = tokenize(s, separator);
    std::vector<T> result;
    try {
      std::transform(
          tokens.begin(), tokens.end(), std::back_inserter(result),
          [](std::string const &s) -> T { return boost::lexical_cast<T>(s); });
    } catch (boost::bad_lexical_cast const &err) {
      error_helper<ErrorParams>(boost::diagnostic_information(err));
    }
    return result;
  }
} // namespace LibLSS

#endif
