/*+
+*/
#pragma once
#include <algorithm>

namespace LibLSS {
  template <typename T>
  class SimpleRange {
    const T begin_;
    const T end_;

  public:
    template <typename RangeType>
    SimpleRange(RangeType &&range, T padding_left = 0, T padding_right = 0)
        : begin_(range.begin() - padding_left),
          end_(range.end() + padding_right) {}
    SimpleRange(T begin, T end) : begin_(begin), end_(end) {}
    T begin() const { return begin_; }
    T end() const { return end_; }
  };
  template <typename RangeType>
  SimpleRange(
      RangeType &&range, decltype(range.begin()) padding_left = 0,
      decltype(range.begin()) padding_right = 0)
      -> SimpleRange<decltype(range.begin())>;

  template <typename T>
  SimpleRange<T> squeeze(SimpleRange<T> &&range, T begin, T end) {
    return SimpleRange(
        std::max(range.begin(), begin), std::min(range.end(), end));
  }
} // namespace LibLSS

// ARES TAG: num_authors = 1
// ARES TAG: name(0) = Ewoud Wempe
// ARES TAG: year(0) = 2022
// ARES TAG: email(0) = wempe@astro.rug.nl
