// This code is derived from GADGET4 peano.cc  developed by Volker Springel.
// Copyright (C) 2014-2020 by Volker Springel
// (vspringel@mpa-garching.mpg.de) and all contributing authors.
// Licensed using GPL-v3
#include <cstdint>
#include <array>

namespace LibLSS {

  typedef uint32_t PeanoPosType;
  static const unsigned char PeanoBitsForPosition = 32;

  struct PeanoKey {
  public:
    PeanoPosType hs, is,
        ls; /* 'hs'-high significance, 'is'-intermediate, 'ls'-low significance bits */

    inline bool operator>=(PeanoKey const &b) const {
      if (hs == b.hs) {
        if (is == b.is) {
          return (ls >= b.ls);
        } else {
          return (is >= b.is);
        }
      } else {
        return hs >= b.hs;
      }
    }

    inline bool operator<(PeanoKey const &b) const {
      if (hs == b.hs) {
        if (is == b.is) {
          return (ls < b.ls);
        } else {
          return (is < b.is);
        }
      } else {
        return hs < b.hs;
      }
    }

    inline PeanoKey operator+(PeanoKey const &b) const {
      PeanoKey c;

      c.ls = ls + b.ls;
      c.is = is + b.is;
      c.hs = hs + b.hs;

      if (c.is < is || c.is < b.is) /* overflow has occurred */
      {
        c.hs += 1;
      }

      if (c.ls < ls || c.ls < b.ls) /* overflow has occurred */
      {
        c.is += 1;
        if (c.is == 0) /* overflown again */
          c.hs += 1;
      }

      /* note: for hs we don't check for overflow explicitly as this would not be represented in the type anyhow */

      return c;
    }

    inline PeanoKey getOffset(
        unsigned int j,
        int bits) /* this returns the peanokey for which  j << bits */
    {
      PeanoKey key = {j, j, j};

      if (bits < PeanoBitsForPosition)
        key.ls <<= bits;
      else
        key.ls = 0;

      int is_bits = bits - PeanoBitsForPosition;

      if (is_bits <= -PeanoBitsForPosition)
        key.is = 0;
      else if (is_bits <= 0)
        key.is >>= -is_bits;
      else if (is_bits < PeanoBitsForPosition)
        key.is <<= is_bits;
      else
        key.is = 0;
      int hs_bits = bits - 2 * PeanoBitsForPosition;

      if (hs_bits <= -PeanoBitsForPosition)
        key.hs = 0;
      else if (hs_bits <= 0)
        key.hs >>= -hs_bits;
      else if (hs_bits < PeanoBitsForPosition)
        key.hs <<= hs_bits;
      else
        key.hs = 0;

      return key;
    }
  };

  PeanoKey
  getPeanoHilbertKey(PeanoPosType x, PeanoPosType y, PeanoPosType z, int bits);
  unsigned char
  getPeanoIncrementalKey(unsigned char pix, unsigned char *rotation);
  std::array<PeanoPosType, 3> inversePeanoHilbertKey(PeanoKey key, int bits);

} // namespace LibLSS