/*+
    ARES/HADES/BORG Package -- ./libLSS/tools/string_tools.hpp
    Copyright (C) 2014-2022 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2022 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2018-2021)

+*/
#ifndef __LIBLSS_STRING_TOOLS_HPP
#define __LIBLSS_STRING_TOOLS_HPP

#include <string>
#include <iostream>
#include <algorithm>
#include <sstream>
#include <vector>
#include <boost/format.hpp>
#include <boost/multi_array.hpp>
#include "libLSS/tools/is_stl_container.hpp"

namespace LibLSS {

  template <typename T>
  typename std::enable_if<is_stl_container_like<T>::value, std::string>::type
  to_string(T const &V) {
    std::ostringstream s;

    std::copy(
        V.begin(), V.end(),
        std::ostream_iterator<typename T::value_type>(s, ","));
    return s.str();
  }

  template <typename T>
  inline typename std::enable_if<
      !is_stl_container_like<T>::value, std::string>::type
  to_string(T const &V) {
    return std::to_string(V);
  }

  static inline std::string to_string(char const *s) { return std::string(s); }

  std::vector<std::string>
  tokenize(std::string const &in, std::string const &separator);

  void multiline(
      std::vector<std::string> &multiline, std::string new_content,
      std::string prefix, int maxSize = 70);

  namespace lssfmt {

    namespace format_detail {

      inline void _format_expansion(boost::format &f) {}

      template <typename A, typename... U>
      inline void _format_expansion(boost::format &f, A &&a, U &&...u) {
        _format_expansion(f % a, u...);
      }

      template <typename... U>
      std::string format(std::string const &s, U &&...args) {
        boost::format f(s);
        _format_expansion(f, std::forward<U>(args)...);
        return boost::str(f);
      }

    } // namespace format_detail

    using format_detail::format;

  } // namespace lssfmt
  /*  template<typename T>
  std::string to_string(boost::multi_array_ref<T,1> const& V) {
    std::ostringstream s;

    std::copy(V.begin(), V.end(), std::ostream_iterator<T>(s, ","));
    return s.str();
  }*/
} // namespace LibLSS

namespace std {
  template <typename Ch, typename Tr, typename T, size_t N>
  std::basic_ostream<Ch, Tr> &
  operator<<(std::basic_ostream<Ch, Tr> &os, std::array<T, N> const &arr) {
    for (size_t i = 0; i < N; i++)
      os << arr[i] << ",";
    return os;
  }
} // namespace std

#endif
