# +
#   ARES/HADES/BORG Package -- ./python/aquila_borg/borg_native.py
#   Copyright (C) 2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
#
#   Additional contributions from:
#      Guilhem Lavaux <guilhem.lavaux@iap.fr> (2020)
#
# +
try:
    from borg_embed import *

    EMBEDDED = True
except ModuleNotFoundError as e:
    # If we do not detect the embedded module we add some
    # symbols to account for the missing decorator functions
    # That way the python code are working both embedded or
    # in standalone python.
    from ._borg import *
    from . import _borg
    import importlib.resources as rc
    import sys
    import copy
    import warnings

    def _unavailable_function():
        raise NotImplementedError("This is not running as part of Hades-Python")

    def _warning_function(f):
        warnings.warn("not embedded borg", RuntimeWarning)
        return f

    if sys.version_info >= (3, 9):
        path_data = rc.files("aquila_borg") / "_data"
    else:
        # Hack the path from the current module file path
        import os
        from pathlib import Path

        path_data = Path(os.path.dirname(__file__)) / "_data"

    if path_data.exists():
        _borg.console().print_std("Using bundled data in python package")
        _borg._setDataDirectory(str(path_data))
    else:
        _borg.console().print_std("Attempting default data path")

    registerLikelihoodBuilder = _warning_function
    registerGravityBuilder = _warning_function
    registerSamplerBuilder = _warning_function
    densityBuilder = _warning_function
    getIniConfigurationFilename = copy.copy(_unavailable_function)
    getIniConfigurationFilename.__doc__ = """Returns the full path of the ini configuration file provided to the main program.

Returns:
   str: full path
    """

    EMBEDDED = False
    del _warning_function

__all__ = [
    "cosmo",
    "forward",
    "bias",
    "likelihood",
    "samplers",
    "domains",
    "modelio",
    "filters",
    "Console",
    "console",
    "version",
    "memoryReport",
    "timingReport",
    "clearTimingReport",
    "clearMemoryReport",
    "registerLikelihoodBuilder",
    "registerGravityBuilder",
    "registerSamplerBuilder",
    "densityBuilder",
    "getIniConfigurationFilename",
    "EMBEDDED",
]


from deprecated import deprecated

deprecated_class = [
    "BorgLpt",
    "Borg2Lpt",
    "BorgPM",
    "BorgLptOpenMP",
    "HadesLog",
    "HadesLinear",
    "EisensteinHu",
]
for c in deprecated_class:
    if not hasattr(forward.models, c):
        continue
    setattr(
        forward.models,
        c,
        deprecated(
            "The tailored binding of native forward models are being deprecated in favor of the "
            "automatically generated bindings in model_lib",
            getattr(forward.models, c),
        ),
    )

# ARES TAG: authors_num = 1
# ARES TAG: name(0) = Guilhem Lavaux
# ARES TAG: year(0) = 2020
# ARES TAG: email(0) = guilhem.lavaux@iap.fr
